package fe

import (
	"bytes"
	"flag"
	"fmt"
	"html/template"
	"io"
	"log"
	"net"
	"net/http"
	"net/url"
	"path/filepath"
	"strconv"
	"strings"
	"time"

	_ "net/http/pprof"

	"git.autistici.org/ale/autoradio"
	"git.autistici.org/ale/autoradio/instrumentation"
	"github.com/PuerkitoBio/ghost/handlers"
)

var (
	proxyStreams = flag.Bool("enable-icecast-proxy", false, "Proxy the local icecast")

	httpStatusCodes   = instrumentation.NewCounter("http.status")
	httpTargetStats   = instrumentation.NewCounter("http.target")
	sourceConnections = instrumentation.NewCounter("http.source_connections")
	source404Errors   = instrumentation.NewCounter("http.source_404")
	sourceErrors      = instrumentation.NewCounter("http.source_errors")
)

type statsResponseWriter struct {
	http.ResponseWriter
	code int
}

func (w *statsResponseWriter) WriteHeader(code int) {
	w.code = code
	w.ResponseWriter.WriteHeader(code)
}

func statsHandler(h http.Handler) http.HandlerFunc {
	return func(w http.ResponseWriter, r *http.Request) {
		wrapw := &statsResponseWriter{w, 200}
		h.ServeHTTP(wrapw, r)
		httpStatusCodes.IncrVar(strconv.Itoa(wrapw.code))
	}
}

// HTTP redirector.
type HttpRedirector struct {
	domain   string
	lb       LoadBalancingPolicy
	client   *autoradio.RadioAPI
	template *template.Template
}

func NewHttpRedirector(client *autoradio.RadioAPI, domain string, lbpolicy string) *HttpRedirector {
	return &HttpRedirector{
		client: client,
		domain: domain,
		lb:     getNamedLoadBalancingPolicy(lbpolicy),
	}
}

// Return an active node, chosen according to the current load
// balancing policy.
func (h *HttpRedirector) pickActiveNode() string {
	nodes, _ := h.client.GetNodes()
	if nodes == nil {
		return ""
	}

	// Filter nodes where Icecast is reported to be up.
	okNodes := make([]*autoradio.NodeStatus, 0, len(nodes))
	for _, n := range nodes {
		if n.IcecastUp {
			okNodes = append(okNodes, n)
		}
	}
	if len(okNodes) == 0 {
		return ""
	}

	result := h.lb.GetNode(okNodes)
	if result == nil {
		return ""
	}
	return result.IP
}

func icecastAddr(server string) string {
	return net.JoinHostPort(server, strconv.Itoa(autoradio.IcecastPort))
}

func streamUrl(server, mountName string) string {
	var serverAddr string
	if *proxyStreams {
		serverAddr = server
	} else {
		serverAddr = icecastAddr(server)
	}
	return fmt.Sprintf("http://%s%s", serverAddr, autoradio.MountNameToIcecastPath(mountName))
}

// Request wrapper that passes a Mount along with the HTTP request.
func (h *HttpRedirector) withMount(f func(*autoradio.Mount, http.ResponseWriter, *http.Request)) http.Handler {
	return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		mountPath := strings.TrimSuffix(r.URL.Path, ".m3u")
		mount, err := h.client.GetMount(mountPath)
		if err != nil {
			http.Error(w, "Not Found", http.StatusNotFound)
			return
		}
		f(mount, w, r)
	})
}

// Serve a M3U response. This simply points back at the stream
// redirect handler.
func (h *HttpRedirector) serveM3U(mount *autoradio.Mount, w http.ResponseWriter, r *http.Request) {
	m3u := strings.TrimSuffix(r.URL.String(), ".m3u") + "\n"
	w.Header().Set("Content-Length", strconv.Itoa(len(m3u)))
	w.Header().Set("Content-Type", "audio/x-mpegurl")
	addDefaultHeaders(w)
	io.WriteString(w, m3u)
}

// Serve a response for a client connection to a relay.
func (h *HttpRedirector) serveRelay(mount *autoradio.Mount, w http.ResponseWriter, r *http.Request) {
	// Find an active node.
	relayAddr := h.pickActiveNode()
	if relayAddr == "" {
		http.Error(w, "No active nodes", http.StatusServiceUnavailable)
		return
	}
	httpTargetStats.IncrVar(ipToMetric(relayAddr))

	// See if we need to serve a M3U response or a redirect.
	if strings.HasSuffix(r.URL.Path, ".m3u") {
		h.serveM3U(mount, w, r)
	} else {
		targetURL := streamUrl(relayAddr, mount.Name)
		http.Redirect(w, r, targetURL, 302)
	}
}

// Handle SOURCE requests.
func (h *HttpRedirector) serveSource(mount *autoradio.Mount, w http.ResponseWriter, r *http.Request) {
	if mount.IsRelay() {
		log.Printf("source: connection to relay stream %s", mount.Name)
		http.Error(w, "Stream is relayed", http.StatusBadRequest)
		source404Errors.Incr()
		return
	}

	sourceConnections.Incr()

	// Find the current master node.
	masterAddr, err := h.client.GetMasterAddr()
	if err != nil {
		log.Printf("source error: no master: %v", err)
		http.Error(w, err.Error(), http.StatusServiceUnavailable)
		sourceErrors.Incr()
		return
	}

	// Create a ReverseProxy on the fly with the right backend
	// address.
	proxy := &ReverseProxy{
		Director: func(req *http.Request) {
			req.URL.Scheme = "http"
			req.URL.Host = icecastAddr(masterAddr)
			req.URL.Path = autoradio.MountNameToIcecastPath(mount.Name)
		},
		FlushInterval: 500 * time.Millisecond,
	}
	proxy.ServeHTTP(w, r)
}

// Serve our cluster status page.
func (h *HttpRedirector) serveStatusPage(w http.ResponseWriter, r *http.Request) {
	nodes, _ := h.client.GetNodes()
	mounts, _ := h.client.ListMounts()
	ctx := struct {
		Domain string
		Nodes  []*autoradio.NodeStatus
		Mounts []*autoradio.Mount
	}{h.domain, nodes, mounts}

	var buf bytes.Buffer
	if err := h.template.ExecuteTemplate(&buf, "index.html", ctx); err != nil {
		http.Error(w, err.Error(), http.StatusInternalServerError)
		return
	}
	w.Header().Set("Content-Type", "text/html; charset=utf-8")
	w.Header().Set("Content-Length", strconv.Itoa(buf.Len()))
	addDefaultHeaders(w)
	w.Write(buf.Bytes())
}

// Run starts the HTTP server on the given addr. Does not return.
func (h *HttpRedirector) Run(addr, staticDir, templateDir string) {
	h.template = template.Must(
		template.ParseGlob(
			filepath.Join(templateDir, "*.html")))

	// Create our HTTP handler stack.
	mux := http.NewServeMux()

	// Serve static content. Responses support gzip encoding.
	mux.Handle(
		"/static/",
		handlers.GZIPHandler(
			http.StripPrefix(
				"/static/",
				http.FileServer(http.Dir(staticDir))),
			nil))

	// Pass /debug/ to the default ServeMux.
	mux.Handle("/debug/", http.DefaultServeMux)

	// The / handler should discriminate between a request for a
	// stream or the status page.
	statush := handlers.GZIPHandler(http.HandlerFunc(h.serveStatusPage), nil)
	relayh := h.withMount(h.serveRelay)
	mux.HandleFunc("/", func(w http.ResponseWriter, r *http.Request) {
		switch {
		case r.URL.Path == "" || r.URL.Path == "/":
			statush.ServeHTTP(w, r)
		default:
			relayh.ServeHTTP(w, r)
		}
	})

	// Add some handlers to support gzip-encoded responses and
	// request logging.
	logopts := handlers.NewLogOptions(nil, handlers.Lshort)
	logopts.Immediate = true
	wraph := handlers.LogHandler(mux, logopts)
	wraph = statsHandler(wraph)

	// Serve proxied requests bypassing the logging and gzip
	// handlers: since they wrap the ResponseWriter, we would be
	// unable to hijack the underlying connection for proxying.
	sourceHandler := h.withMount(h.serveSource)
	root := http.NewServeMux()

	// Optionally enable a reverse proxy to the local Icecast.
	if *proxyStreams {
		iceurl, _ := url.Parse(fmt.Sprintf("http://localhost:%d", autoradio.IcecastPort))
		root.Handle(autoradio.IcecastMountPrefix+"/", NewSingleHostReverseProxy(iceurl))
	}
	root.HandleFunc("/", func(w http.ResponseWriter, r *http.Request) {
		if r.Method == "SOURCE" {
			sourceHandler.ServeHTTP(w, r)
		} else {
			wraph.ServeHTTP(w, r)
		}
	})

	httpServer := &http.Server{
		Addr:         addr,
		Handler:      root,
		ReadTimeout:  10 * time.Second,
		WriteTimeout: 10 * time.Second,
	}

	log.Printf("starting HTTP server on %s/tcp", httpServer.Addr)
	log.Fatal(httpServer.ListenAndServe())
}

func addDefaultHeaders(w http.ResponseWriter) {
	w.Header().Set("Expires", "-1")
	w.Header().Set("Cache-Control", "private, max-age=0")
}
